/* authorize depending on XML response of external server */
/* See http://www.hoboes.com/Mimsy/?ART=644 for tutorial */
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"

#include "apr_strings.h"
#include "apr_xml.h"

/* create a request to send client information to an authentication server */
static char *makeAuthRequest(request_rec *request) {
	char *authRequest;
	char *remote_ip;
	char *uri;
	
	remote_ip = request->connection->remote_ip;
	uri = request->uri;
	
	authRequest = apr_pstrcat(request->pool, "GET /authorization.php?ip=", remote_ip, "&page=", uri, "\n", NULL);
	
	return authRequest;
}

static int logError(apr_status_t status, char *message, request_rec *request) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, status, request, "external_auth: error %s", message);
	return HTTP_SERVICE_UNAVAILABLE;
}

static int getAuthentication(request_rec *request, char *authRequest, const char **xml_response, apr_size_t *xml_length) {
	apr_socket_t *sock;
	apr_sockaddr_t *sockaddr;
	apr_status_t status;
	//timeout appears to be in microseconds, so time out after 5 seconds
	apr_interval_time_t timeout = 5000000;
	apr_size_t item_size;
	apr_bucket*	xmlBucket;

	if ((status = apr_sockaddr_info_get(&sockaddr, "216.92.252.156", APR_INET, 80, 0, request->pool)) != APR_SUCCESS) {
		return logError(status, "creating socket address", request);
	}

	if ((status = apr_socket_create(&sock, sockaddr->family, SOCK_STREAM, APR_PROTO_TCP, request->pool)) != APR_SUCCESS) {
		return logError(status, "creating socket", request);
	}
		
	if ((status = apr_socket_timeout_set(sock, timeout)) != APR_SUCCESS) {
		return logError(status, "setting socket timeout", request);
	}

	if ((status = apr_socket_connect(sock, sockaddr)) != APR_SUCCESS) {
		return logError(status, "connecting", request);
	}
	
	item_size = strlen(authRequest);
	if ((status = apr_socket_send(sock, authRequest, &item_size)) != APR_SUCCESS) {
		return logError(status, "sending query", request);
	}
	
	xmlBucket = apr_bucket_socket_create(sock, request->connection->bucket_alloc);
	if ((status = apr_bucket_read(xmlBucket, xml_response, xml_length, APR_BLOCK_READ)) != APR_SUCCESS) {
		return logError(status, "receiving response", request);
	}
	
	apr_socket_close(sock);
	apr_bucket_destroy(xmlBucket);
	
	return APR_SUCCESS;
}


static int parseAuthentication(request_rec *request, const char *xml, apr_size_t xml_length, apr_table_t **userInfo) {
	apr_xml_parser *parser;
	apr_status_t status;
	apr_xml_doc *parsedDocument;
	const char *elementValue;
	apr_size_t elementSize;

	apr_xml_elem *element;

	parser = apr_xml_parser_create(request->pool);

	if ((status = apr_xml_parser_feed(parser, xml, xml_length)) == APR_SUCCESS) {
		if ((status = apr_xml_parser_done(parser, &parsedDocument)) == APR_SUCCESS) {
			*userInfo = apr_table_make(request->pool, 0);
			for (element = parsedDocument->root->first_child; element != NULL; element = element->next) {
				apr_xml_to_text(request->pool, element, APR_XML_X2T_INNER, NULL, NULL, &elementValue, &elementSize);
				apr_table_set(*userInfo, element->name, elementValue);
			}
			return APR_SUCCESS;
		} else {
			return logError(status, "parsing xml", request);
		}
	} else {
		return logError(status, "feeding xml", request);
	}

}


static int checkAuthentication(request_rec *request, apr_table_t *userInfo) {
	const char *authorization;

	if (authorization = apr_table_get(userInfo, "authorization")) {
		if (strcmp(authorization, "Yes") == 0) {
			apr_table_overlap(request->subprocess_env, userInfo, APR_OVERLAP_TABLES_MERGE);
			return OK;
		} else if (strcmp(authorization, "No") == 0) {
			return HTTP_FORBIDDEN;
		} else {
			return logError(0, "authorization not valid", request);
		}
	} else {
		return logError(0, "no authorization found", request);
	}
}


static int externalAuthorizer(request_rec *request) {
	const char *current_auth;
	char *authRequest;
	const char *xml_response;
	apr_size_t xml_length;
	int status;
	apr_table_t *userInfo;

	//if this is not a HOBOES auth type, bail and let someone else handle it
	current_auth = ap_auth_type(request);
	if (!current_auth || strcasecmp(current_auth, "hoboes")) {
		return DECLINED;
	}

	//create authorization request
	authRequest = makeAuthRequest(request);

	//make call to authentication server
	if ((status = getAuthentication(request, authRequest, &xml_response, &xml_length)) != APR_SUCCESS) {
		return status;
	}

	//parse authentication response
	if ((status = parseAuthentication(request, xml_response, xml_length, &userInfo)) != APR_SUCCESS) {
		return status;
	}

	//check authentication response
	return checkAuthentication(request, userInfo);

}

static void myRegisterHooks(apr_pool_t *p) {
	ap_hook_check_user_id(externalAuthorizer, NULL, NULL, APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA external_auth_module = {
	STANDARD20_MODULE_STUFF,
	NULL, /* create per-directory config structures */
	NULL, /* merge per-directory config structures */
	NULL, /* create per-server config structures */
	NULL, /* merge per-server config structures */
	NULL, /* command handlers */
	myRegisterHooks /* register hooks */
};