/* authorize depending on XML response of external server */
/* See http://www.hoboes.com/Mimsy/?ART=644 for tutorial */
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"

#include "apr_strings.h"
#include "apr_xml.h"

#include "math.h"

char *authHost = "www.hoboes.com";
#define HEADEREND CRLF CRLF

/* create a request to send client information to an authentication server */
static char *makeAuthRequest(request_rec *request) {
	char *authRequest;
	char *remote_ip;
	char *uri;
	
	remote_ip = request->connection->remote_ip;
	uri = request->uri;
	
	authRequest = apr_pstrcat(request->pool, "GET /authorization.php?ip=", remote_ip, "&page=", uri, " HTTP/1.1", CRLF, "Host: ", authHost, CRLF, "Content-Length: 0", CRLF, "Connection: close", HEADEREND, NULL);
	
	return authRequest;
}

static int logError(apr_status_t status, char *message, request_rec *request) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, status, request, "external_auth: error %s", message);
	return HTTP_SERVICE_UNAVAILABLE;
}

static void logNotice(char *title, char *message, request_rec *request) {
	ap_log_rerror(APLOG_MARK, APLOG_NOTICE, 0, request, "%s: %s", title, message);
}

static void logInteger(char *title, int value, request_rec *request) {
	logNotice(title, apr_itoa(request->pool, value), request);
}

//unchunk an HTTP 1.1 response if necessary
static int parseResponse(const char **xml_response, apr_size_t *xml_length, request_rec *request) {
	//remember the start of headers
	const char *headerStart = *xml_response;
	apr_status_t status = APR_SUCCESS;

	//get past the headers to the beginning of the body (and the first chunk size)
	char *chunkSizeStart = strstr(*xml_response, HEADEREND)+strlen(HEADEREND);
	
	//look for the presence of chunking
	//this only works because we expect short bodies with well-formed XML
	//HTTP headers are not case sensitive
	if (strcasestr(headerStart, CRLF "transfer-encoding: chunked" CRLF)) {
		char *chunkStart;
		int chunkLength;
		int chunkLengthCounter;
		char *chunkLengthPointer;

		int chunkLengthPart;

		//the good data will be less than the size of headers+body+chunksizes
		char *goodResponse = malloc(*xml_length);
		int fullLength = 0;

		do {
			//get the chunk length: all chunk lengths begin and end with CRLF
			chunkStart = strstr(chunkSizeStart, CRLF);
			
			//work the chunk size hex string backwards to convert from hex to integer
			chunkLengthPointer = chunkStart-1;
			chunkLength = 0;
			chunkLengthCounter = 0;
			while (chunkLengthPointer >= chunkSizeStart) {
				chunkLengthPart = *chunkLengthPointer;
				if (chunkLengthPart >= 97 && chunkLengthPart <= 102) {
					//letter a - f
					chunkLengthPart -= 97;
					chunkLengthPart += 10;
				} else if (chunkLengthPart >=48 && chunkLengthPart <=57) {
					//number 0 - 9
					chunkLengthPart -= 48;
				} else {
					logInteger("bad hex character", chunkLengthPart, request);
					return HTTP_SERVICE_UNAVAILABLE;
				}
				chunkLength += chunkLengthPart*pow(16, chunkLengthCounter);
				chunkLengthCounter++;
				chunkLengthPointer -= 1;
			}
			
			//copy the chunk to the good data
			if (chunkLength>0) {
				//logInteger("chunk length", chunkLength, request);
				chunkStart += strlen(CRLF);
				//logNotice("chunk", chunkStart, request);
				strncpy(goodResponse+fullLength, chunkStart, chunkLength);
				fullLength += chunkLength;
				
				//the next chunk size string starts where this chunk ends
				chunkSizeStart = chunkStart + chunkLength; // + strlen(CRLF);
			}
		} while (chunkLength > 0);

		*xml_length = fullLength;
		*xml_response = goodResponse;

		//logInteger("unchunked length", fullLength, request);
	} else {
		*xml_response = chunkSizeStart;
		*xml_length = strlen(*xml_response);

		//logInteger("no need to unchunk length", *xml_length, request);
	}
	
	return status;
}

//read from an existing connection into a bucket brigade and then flatten it into xml_response
static int readResponse(request_rec *request, apr_socket_t *sock, const char **xml_response, apr_size_t *xml_length) {
	int connected;
	apr_size_t runningLength;
	apr_status_t status = APR_SUCCESS;

	apr_bucket *xmlBucket;
	apr_bucket_brigade *xmlBrigade;

	xmlBrigade = apr_brigade_create(request->pool, request->connection->bucket_alloc);
	xmlBucket = apr_bucket_socket_create(sock, request->connection->bucket_alloc);
	
	connected = 1;
	runningLength = 0;
	while (connected) {
		if ((status = apr_bucket_read(xmlBucket, xml_response, xml_length, APR_BLOCK_READ)) != APR_SUCCESS) {
			return logError(status, "receiving response", request);
		}
		
		if (*xml_length <= 0) {
			apr_bucket_destroy(xmlBucket);
			connected = 0;
		} else {
			runningLength += *xml_length;
			APR_BRIGADE_INSERT_TAIL(xmlBrigade, xmlBucket);
		}

		xmlBucket = apr_bucket_socket_create(sock, request->connection->bucket_alloc);
	}

	apr_socket_close(sock);

	*xml_length = runningLength;
	*xml_response = malloc(runningLength);
	if ((status = apr_brigade_flatten(xmlBrigade, (char *) *xml_response, &runningLength)) != APR_SUCCESS) {
		logError(status, "flattening response", request);
	}

	apr_brigade_destroy(xmlBrigade);

	return status;
}


static int getAuthentication(request_rec *request, char *authRequest, const char **xml_response, apr_size_t *xml_length) {
	apr_socket_t *sock;
	apr_sockaddr_t *sockaddr;
	apr_status_t status;
	//timeout appears to be in microseconds, so time out after 5 seconds
	apr_interval_time_t timeout = 5000000;
	apr_size_t item_size;

	if ((status = apr_sockaddr_info_get(&sockaddr, authHost, APR_INET, 80, 0, request->pool)) != APR_SUCCESS) {
		return logError(status, "creating socket address", request);
	}

	if ((status = apr_socket_create(&sock, sockaddr->family, SOCK_STREAM, APR_PROTO_TCP, request->pool)) != APR_SUCCESS) {
		return logError(status, "creating socket", request);
	}
		
	if ((status = apr_socket_timeout_set(sock, timeout)) != APR_SUCCESS) {
		return logError(status, "setting socket timeout", request);
	}

	if ((status = apr_socket_connect(sock, sockaddr)) != APR_SUCCESS) {
		return logError(status, "connecting", request);
	}
	
	item_size = strlen(authRequest);
	if ((status = apr_socket_send(sock, authRequest, &item_size)) != APR_SUCCESS) {
		return logError(status, "sending query", request);
	}

	if ((status = readResponse(request, sock, xml_response, xml_length)) != APR_SUCCESS) {
		logNotice("authRequest", authRequest, request);
		return logError(status, "receiving XML response", request);
	}
	
	if ((status = parseResponse(xml_response, xml_length, request)) != APR_SUCCESS) {
		return logError(status, "unchunking XML response", request);
	}

	return APR_SUCCESS;
}

static int parseAuthentication(request_rec *request, const char *xml, apr_size_t xml_length, apr_table_t **userInfo) {
	apr_xml_parser *parser;
	apr_status_t status;
	apr_xml_doc *parsedDocument;
	const char *elementValue;
	apr_size_t elementSize;

	apr_xml_elem *element;

	parser = apr_xml_parser_create(request->pool);

	if ((status = apr_xml_parser_feed(parser, xml, xml_length)) == APR_SUCCESS) {
		if ((status = apr_xml_parser_done(parser, &parsedDocument)) == APR_SUCCESS) {
			*userInfo = apr_table_make(request->pool, 0);
			for (element = parsedDocument->root->first_child; element != NULL; element = element->next) {
				apr_xml_to_text(request->pool, element, APR_XML_X2T_INNER, NULL, NULL, &elementValue, &elementSize);
				apr_table_set(*userInfo, element->name, elementValue);
			}
			return APR_SUCCESS;
		} else {
			return logError(status, "parsing xml", request);
		}
	} else {
		logNotice("xml", (char *)xml, request);
		return logError(status, "feeding xml", request);
	}

}

static int checkAuthentication(request_rec *request, apr_table_t *userInfo) {
	const char *authorization;

	if (authorization = apr_table_get(userInfo, "authorization")) {
		if (strcmp(authorization, "Yes") == 0) {
			apr_table_overlap(request->subprocess_env, userInfo, APR_OVERLAP_TABLES_MERGE);
			return OK;
		} else if (strcmp(authorization, "No") == 0) {
			return HTTP_FORBIDDEN;
		} else {
			return logError(0, "authorization not valid", request);
		}
	} else {
		return logError(0, "no authorization found", request);
	}
}


static int externalAuthorizer(request_rec *request) {
	const char *current_auth;
	char *authRequest;
	const char *xml_response;
	apr_size_t xml_length;
	int status;
	apr_table_t *userInfo;

	//if this is not a HOBOES auth type, bail and let someone else handle it
	current_auth = ap_auth_type(request);
	if (!current_auth || strcasecmp(current_auth, "hoboes")) {
		return DECLINED;
	}

	//create authorization request
	authRequest = makeAuthRequest(request);

	//make call to authentication server
	if ((status = getAuthentication(request, authRequest, &xml_response, &xml_length)) != APR_SUCCESS) {
		return status;
	}

	//parse authentication response
	if ((status = parseAuthentication(request, xml_response, xml_length, &userInfo)) != APR_SUCCESS) {
		return status;
	}

	//check authentication response
	return checkAuthentication(request, userInfo);

}

static void myRegisterHooks(apr_pool_t *p) {
	ap_hook_check_user_id(externalAuthorizer, NULL, NULL, APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA external_auth_module = {
	STANDARD20_MODULE_STUFF,
	NULL, /* create per-directory config structures */
	NULL, /* merge per-directory config structures */
	NULL, /* create per-server config structures */
	NULL, /* merge per-server config structures */
	NULL, /* command handlers */
	myRegisterHooks /* register hooks */
};